#include "pch.h"
#include "D3D11Texture.h"
#include "Engine.h"
#include "D3D11GraphicsEngine.h"
#include "GothicAPI.h"
#include <D3DX11.h>

D3D11Texture::D3D11Texture(void)
{
	Texture = NULL;
	ShaderResourceView = NULL;
}

D3D11Texture::~D3D11Texture(void)
{
	if (Texture)Texture->Release();
	if (ShaderResourceView)ShaderResourceView->Release();
}

/** Initializes the texture object */
XRESULT D3D11Texture::Init(INT2 size, ETextureFormat format, UINT mipMapCount, void* data, const std::string& fileName)
{
	HRESULT hr;
	D3D11GraphicsEngine* engine = (D3D11GraphicsEngine *)Engine::GraphicsEngine;

	Engine::GAPI->EnterResourceCriticalSection();

	TextureFormat = (DXGI_FORMAT)format;
	TextureSize = size;

	CD3D11_TEXTURE2D_DESC textureDesc(
		(DXGI_FORMAT)format,
		size.x,
		size.y,
		1,
		mipMapCount,
		D3D11_BIND_SHADER_RESOURCE, D3D11_USAGE_DEFAULT, 0, 1, 0, 0);

	LE(engine->GetDevice()->CreateTexture2D(&textureDesc, nullptr, &Texture));

#ifndef PUBLIC_RELEASE
	Texture->SetPrivateData(WKPDID_D3DDebugObjectName, fileName.size(), fileName.c_str());
#endif

	D3D11_SHADER_RESOURCE_VIEW_DESC descRV;
	ZeroMemory(&descRV, sizeof(descRV));
	descRV.Format = DXGI_FORMAT_UNKNOWN;
	descRV.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
	descRV.Texture2D.MipLevels = mipMapCount;
	descRV.Texture2D.MostDetailedMip = 0;
	LE(engine->GetDevice()->CreateShaderResourceView(Texture, &descRV, &ShaderResourceView));

	Engine::GAPI->LeaveResourceCriticalSection();

	return XR_SUCCESS;
}

/** Initializes the texture from a file */
XRESULT D3D11Texture::Init(const std::string& file)
{
	HRESULT hr;
	D3D11GraphicsEngine* engine = (D3D11GraphicsEngine *)Engine::GraphicsEngine;

	LogInfo() << "Loading Engine-Texture: " << file;

	LE(D3DX11CreateShaderResourceViewFromFileA(engine->GetDevice(), file.c_str(), NULL, NULL, &ShaderResourceView, NULL));

	if(!ShaderResourceView)
		return XR_FAILED;
	
	ID3D11Texture2D* res;
	ShaderResourceView->GetResource((ID3D11Resource **)&res);

	D3D11_TEXTURE2D_DESC desc;
	res->GetDesc(&desc);

	Texture = res;
	TextureFormat = desc.Format;

	TextureSize.x = desc.Width;
	TextureSize.y = desc.Height;

	return XR_SUCCESS;
}

/** Updates the Texture-Object */
XRESULT D3D11Texture::UpdateData(void* data, int mip)
{
	D3D11GraphicsEngine* engine = (D3D11GraphicsEngine *)Engine::GraphicsEngine;

	// Enter the critical section for safety while executing the deferred command list
	Engine::GAPI->EnterResourceCriticalSection();
	engine->GetContext()->UpdateSubresource(Texture, mip, NULL, data, GetRowPitchBytes(mip), GetSizeInBytes(mip));
	Engine::GAPI->LeaveResourceCriticalSection();

	return XR_SUCCESS;
}

/** Updates the Texture-Object using the deferred context (For loading in an other thread) */
XRESULT D3D11Texture::UpdateDataDeferred(void* data, int mip)
{
	D3D11GraphicsEngine* engine = (D3D11GraphicsEngine *)Engine::GraphicsEngine;

	// Enter the critical section for safety while executing the deferred command list
	Engine::GAPI->EnterResourceCriticalSection();
	engine->GetDeferredContext()->UpdateSubresource(Texture, mip, NULL, data, GetRowPitchBytes(mip), GetSizeInBytes(mip));
	Engine::GAPI->LeaveResourceCriticalSection();

	return XR_SUCCESS;
}

/** Returns the RowPitch-Bytes */
UINT D3D11Texture::GetRowPitchBytes(int mip)
{
	int px = (int)std::max(1.0f, floor(TextureSize.x / pow(2.0f, mip)));
	int py = (int)std::max(1.0f, floor(TextureSize.y / pow(2.0f, mip)));
	//int px = TextureSize.x;
	//int py = TextureSize.y;

	if (TextureFormat == DXGI_FORMAT_BC1_UNORM || TextureFormat == DXGI_FORMAT_BC2_UNORM ||
		TextureFormat == DXGI_FORMAT_BC3_UNORM)
	{
		return Toolbox::GetDDSRowPitchSize(px, TextureFormat == DXGI_FORMAT_BC1_UNORM);
	}
	else // Use R8G8B8A8
	{
		return px * 4;
	}
}

/** Returns the size of the texture in bytes */
UINT D3D11Texture::GetSizeInBytes(int mip)
{
	int px = (int)std::max(1.0f, floor(TextureSize.x / pow(2.0f, mip)));
	int py = (int)std::max(1.0f, floor(TextureSize.y / pow(2.0f, mip)));
	//int px = TextureSize.x;
	//int py = TextureSize.y;

	if (TextureFormat == DXGI_FORMAT_BC1_UNORM || TextureFormat == DXGI_FORMAT_BC2_UNORM ||
		TextureFormat == DXGI_FORMAT_BC3_UNORM)
	{
		return Toolbox::GetDDSStorageRequirements(px, py, TextureFormat == DXGI_FORMAT_BC1_UNORM);
	}
	else // Use R8G8B8A8
	{
		return px * py * 4;
	}
}

/** Binds this texture to a pixelshader */
XRESULT D3D11Texture::BindToPixelShader(int slot)
{
	D3D11GraphicsEngine* engine = (D3D11GraphicsEngine *)Engine::GraphicsEngine;

	ID3D11ShaderResourceView* srv = ShaderResourceView;
	engine->GetContext()->PSSetShaderResources(slot, 1, &srv);

	return XR_SUCCESS;
}

/** Binds this texture to a pixelshader */
XRESULT D3D11Texture::BindToVertexShader(int slot)
{
	D3D11GraphicsEngine* engine = (D3D11GraphicsEngine *)Engine::GraphicsEngine;

	ID3D11ShaderResourceView* srv = ShaderResourceView;
	engine->GetContext()->VSSetShaderResources(slot, 1, &srv);

	return XR_SUCCESS;
}
