#include "pch.h"
#include "GVegetationBox.h"
#include "GMeshSimple.h"
#include "Engine.h"
#include "GothicAPI.h"
#include "zCBspTree.h"
#include "BaseGraphicsEngine.h"
#include "BaseLineRenderer.h"
#include "BaseTexture.h"
#include "D3D11GraphicsEngine.h"
#include "zCMaterial.h"

GVegetationBox::GVegetationBox(void)
{
	VegetationMesh = NULL;
	VegetationTexture = NULL;
	InstancingBuffer = NULL;
	GrassCB = NULL;
	MeshTexture = NULL;
	DrawBoundingBox = false;
}


GVegetationBox::~GVegetationBox(void)
{
	delete VegetationMesh;
	delete InstancingBuffer;
	delete VegetationTexture;
	delete GrassCB;
}

/** Returns true if the given position is inside the box */
bool GVegetationBox::PositionInsideBox(const D3DXVECTOR3& p)
{
	if( p.x > BoxMin.x &&
		p.y > BoxMin.y &&
		p.z > BoxMin.z &&
		p.x < BoxMax.x &&
		p.y < BoxMax.y &&
		p.z < BoxMax.z )
		return true;

	return false;
}

XRESULT GVegetationBox::InitVegetationBox(  MeshInfo* mesh, 
								const std::string& vegetationMesh,
								float density, 
								float maxSize,
								zCTexture* meshTexture)
{
	if(VegetationMesh)
	{
		LogWarn() << "Tried to init GVegetationBox twice!";
		return XR_FAILED;
	}

	// Load vegetationmesh
	VegetationMesh = new GMeshSimple;
	if(XR_SUCCESS != VegetationMesh->LoadMesh("system\\GD3D11\\Meshes\\grass02.3ds"))
	{
		delete VegetationMesh;
		return XR_FAILED;
	}

	Engine::GraphicsEngine->CreateTexture(&VegetationTexture);
	VegetationTexture->Init("system\\GD3D11\\Meshes\\grass02.png");

	MeshTexture = meshTexture;

	// Compute boundingbox and polys
	BoxMax = D3DXVECTOR3(FLT_MIN, FLT_MIN, FLT_MIN);
	BoxMin = D3DXVECTOR3(FLT_MAX, FLT_MAX, FLT_MAX);

	for(unsigned int i=0;i<mesh->Vertices.size();i++)
	{
		BoxMin.x = BoxMin.x > mesh->Vertices[i].Position.x ? mesh->Vertices[i].Position.x : BoxMin.x;
		BoxMin.y = BoxMin.y > mesh->Vertices[i].Position.y ? mesh->Vertices[i].Position.y : BoxMin.y;
		BoxMin.z = BoxMin.z > mesh->Vertices[i].Position.z ? mesh->Vertices[i].Position.z : BoxMin.z;

		BoxMax.x = BoxMax.x < mesh->Vertices[i].Position.x ? mesh->Vertices[i].Position.x : BoxMax.x;
		BoxMax.y = BoxMax.y < mesh->Vertices[i].Position.y ? mesh->Vertices[i].Position.y : BoxMax.y;
		BoxMax.z = BoxMax.z < mesh->Vertices[i].Position.z ? mesh->Vertices[i].Position.z : BoxMax.z;
	}

	std::vector<D3DXVECTOR3> trisInside;
	for(unsigned int i=0;i<mesh->Indices.size();i+=3)
	{
		D3DXVECTOR3 tri[3];

		tri[0] = *mesh->Vertices[mesh->Indices[i]].Position.toD3DXVECTOR3();
		tri[1] = *mesh->Vertices[mesh->Indices[i+1]].Position.toD3DXVECTOR3();
		tri[2] = *mesh->Vertices[mesh->Indices[i+2]].Position.toD3DXVECTOR3();

		trisInside.push_back(tri[0]);
		trisInside.push_back(tri[1]);
		trisInside.push_back(tri[2]);
	}

	InitSpotsRandom(trisInside);

	return XR_SUCCESS;
}

/** Initializes the vegetationbox */
XRESULT GVegetationBox::InitVegetationBox(const D3DXVECTOR3& min, 
										  const D3DXVECTOR3& max, 
										  const std::string& vegetationMesh, 
										  float density, 
										  float maxSize,
										  const std::string& restrictByTexture,
										  EShape shape)
{
	if(VegetationMesh)
	{
		LogWarn() << "Tried to init GVegetationBox twice!";
		return XR_FAILED;
	}

	// Load vegetationmesh
	VegetationMesh = new GMeshSimple;
	if(XR_SUCCESS != VegetationMesh->LoadMesh("system\\GD3D11\\Meshes\\grass02.3ds"))
	{
		delete VegetationMesh;
		return XR_FAILED;
	}

	Engine::GraphicsEngine->CreateTexture(&VegetationTexture);
	VegetationTexture->Init("system\\GD3D11\\Meshes\\grass02.png");

	if(restrictByTexture != "")
	{
		zCMaterial* m = Engine::GAPI->GetMaterialByTextureName(restrictByTexture);
		MeshTexture = m ? m->GetTexture() : NULL;
	}

	BoxMax = max;
	BoxMin = min;

	// Get polygons laying in this box
	zCPolygon** p = Engine::GAPI->GetLoadedWorldInfo()->BspTree->GetPolygons();
	std::vector<D3DXVECTOR3> polysInside;

	// Get polys inside the box //TODO: Get crossing polys too!
	for(int i=0;i<Engine::GAPI->GetLoadedWorldInfo()->BspTree->GetNumPolys();i++)
	{
		for(int v=0;v<4;v++)
		{
			if(v == 4)
			{
				// Check center too
				D3DXVECTOR3 tri[] = {   *p[i]->getVertices()[0]->Position.toD3DXVECTOR3(), 
										*p[i]->getVertices()[1]->Position.toD3DXVECTOR3(), 
										*p[i]->getVertices()[2]->Position.toD3DXVECTOR3()};

				// Get the center
				D3DXVECTOR3 center = (tri[0] + tri[1] + tri[2]) / 3.0f;

				if(PositionInsideBox(*p[i]->getVertices()[v]->Position.toD3DXVECTOR3()))
				{
					// Restrict by texture
					if(restrictByTexture.length() && 
						p[i]->GetMaterial() &&
						p[i]->GetMaterial()->GetTexture() &&
						p[i]->GetMaterial()->GetTexture()->GetNameWithoutExt() == restrictByTexture)
					{
						polysInside.push_back(tri[0]);
						polysInside.push_back(tri[1]);
						polysInside.push_back(tri[2]);
					}else if(!restrictByTexture.length())
					{
						polysInside.push_back(tri[0]);
						polysInside.push_back(tri[1]);
						polysInside.push_back(tri[2]);
					}

					// Use the texture of the first poly we find
					if(!MeshTexture)
					{
						MeshTexture = p[i]->GetMaterial() ? p[i]->GetMaterial()->GetTexture() : NULL;
					}
				}
				break;
			}

			if(PositionInsideBox(*p[i]->getVertices()[v]->Position.toD3DXVECTOR3()))
			{
				D3DXVECTOR3 tri[] = {   *p[i]->getVertices()[0]->Position.toD3DXVECTOR3(), 
										*p[i]->getVertices()[1]->Position.toD3DXVECTOR3(), 
										*p[i]->getVertices()[2]->Position.toD3DXVECTOR3()};

				// Restrict by texture
				if(restrictByTexture.length() && 
					p[i]->GetMaterial() &&
					p[i]->GetMaterial()->GetTexture() &&
					p[i]->GetMaterial()->GetTexture()->GetNameWithoutExt() == restrictByTexture)
				{
					polysInside.push_back(tri[0]);
					polysInside.push_back(tri[1]);
					polysInside.push_back(tri[2]);
				}else if(!restrictByTexture.length())
				{
					polysInside.push_back(tri[0]);
					polysInside.push_back(tri[1]);
					polysInside.push_back(tri[2]);
				}

				// Use the texture of the first poly we find
				if(!MeshTexture)
				{
					MeshTexture = p[i]->GetMaterial() ? p[i]->GetMaterial()->GetTexture() : NULL;
				}
				break;
			}
		}
	}


	InitSpotsRandom(polysInside, shape);

	return XR_SUCCESS;
}

/** Puts trasformation for the given spots */
void GVegetationBox::InitSpotsRandom(const std::vector<D3DXVECTOR3>& trisInside, EShape shape)
{
	D3DXVECTOR3 mid = BoxMin * 0.5f + BoxMax * 0.5f;
	D3DXVECTOR3 bs = (BoxMax - BoxMin);
	float rad = std::min(bs.x, bs.z) / 2.0f;

	// Find random spots on the polygons (TODO: This is still based off the size of the polygons!)
	std::vector<D3DXVECTOR3> spots;
	for(unsigned int i=0;i<trisInside.size();i+=3)
	{
		for(unsigned int d=0;d<30;d++)
		{
			D3DXVECTOR3 tri[] = {   trisInside[i], 
									trisInside[i+1], 
									trisInside[i+2]};

			// Get 2 random points on the edges
			D3DXVECTOR3 rp[3];
			D3DXVec3Lerp(&rp[0], &tri[0], &tri[1], Toolbox::frand());
			D3DXVec3Lerp(&rp[1], &tri[0], &tri[2], Toolbox::frand());

			// Get the last point on that random made edge
			D3DXVec3Lerp(&rp[2], &rp[0], &rp[1], Toolbox::frand());

			if(PositionInsideBox(rp[2]))
			{
				if(shape == S_Circle) // Restrict to smalles circle inside our AABB
				{
					D3DXVECTOR2 m2 = D3DXVECTOR2(mid.x, mid.z);
					float dist = D3DXVec2Length(&(D3DXVECTOR2(rp[2].x, rp[2].z) - m2));

					if(dist >= rad)
						continue;
				}

				spots.push_back(rp[2]);
			}
		}
	}


	// Create the transformation matrices for every spot
	for(unsigned int i=0;i<spots.size();i++)
	{
		D3DXMATRIX w;
		D3DXMATRIX s;
		D3DXMATRIX r;
		D3DXMatrixTranslation(&w, spots[i].x, spots[i].y, spots[i].z);
		float scale = Toolbox::lerp(60, 120, Toolbox::frand());
		D3DXMatrixScaling(&s, scale,scale,scale);
		D3DXMatrixRotationY(&r, Toolbox::frand() * (float)D3DX_PI * 2.0f);

		w = r * s * w;

		D3DXMatrixTranspose(&w, &w);

		VegetationSpots.push_back(w);
	}

	if(VegetationSpots.empty())
	{
		return;
	}

	// Create instancing buffer for this box
	Engine::GraphicsEngine->CreateVertexBuffer(&InstancingBuffer);
	InstancingBuffer->Init(&VegetationSpots[0], VegetationSpots.size() * sizeof(D3DXMATRIX));

	// Create constant buffer
	Engine::GraphicsEngine->CreateConstantBuffer(&GrassCB, NULL, sizeof(GrassConstantBuffer));
	return;
}

/** Draws this vegetation box */
void GVegetationBox::RenderVegetation(const D3DXVECTOR3& eye)
{
	float drawRadius = Engine::GAPI->GetRendererState()->RendererSettings.OutdoorSmallVobDrawRadius;

	float dist = Toolbox::ComputePointAABBDistance(eye, BoxMin, BoxMax);
	if(dist > drawRadius)
		return;

	if(VegetationSpots.empty())
	{
		return;
	}



	if(MeshTexture)
	{	
		if(MeshTexture->CacheIn(0.6f) == zRES_CACHED_IN)
			MeshTexture->Bind(0);
	}

	VegetationTexture->BindToPixelShader(1);

	Engine::GAPI->GetRendererState()->RasterizerState.CullMode = GothicRasterizerStateInfo::CM_CULL_NONE;
	Engine::GAPI->GetRendererState()->RasterizerStateDirty = true;

	// Enable alpha-to-coverage
	
	if(Engine::GAPI->GetRendererState()->RendererSettings.VegetationAlphaToCoverage)
	{
		Engine::GAPI->GetRendererState()->BlendState.SetDefault();
		Engine::GAPI->GetRendererState()->BlendState.BlendEnabled = false;
		Engine::GAPI->GetRendererState()->BlendState.AlphaToCoverage = Engine::GAPI->GetRendererState()->RendererSettings.VegetationAlphaToCoverage;
	}

	Engine::GAPI->GetRendererState()->BlendStateDirty = true;

	Engine::GraphicsEngine->SetActiveVertexShader("VS_GrassInstanced");
	Engine::GraphicsEngine->SetActivePixelShader("PS_Grass");

	((D3D11GraphicsEngine*)Engine::GraphicsEngine)->SetupVS_ExMeshDrawCall();
	((D3D11GraphicsEngine*)Engine::GraphicsEngine)->SetupVS_ExConstantBuffer();
	//((D3D11GraphicsEngine*)Engine::GraphicsEngine)->SetupVS_ExPerInstanceConstantBuffer();

	// Unseed randomizer to always have the same set of scales/rotations
	//srand(0);

	D3DXMATRIX view;
	Engine::GAPI->GetViewMatrix(&view);
	D3DXMatrixTranspose(&view, &view);

	GrassConstantBuffer gcb;
	D3DXVec3TransformNormal(gcb.G_NormalVS.toD3DXVECTOR3(), &D3DXVECTOR3(0,1,0), &view);
	gcb.G_Time = Engine::GAPI->GetTimeSeconds();
	gcb.G_WindStrength = Engine::GAPI->GetRendererState()->RendererSettings.GlobalWindStrength;
	GrassCB->UpdateBuffer(&gcb);
	GrassCB->BindToVertexShader(1);

	// Draw the batch
	VegetationMesh->DrawBatch(InstancingBuffer, VegetationSpots.size(), sizeof(D3DXMATRIX));
	
	/*for(int i=0;i<VegetationSpots.size();i++)
	{
		//float sizeMod = 1 - powf((dist / drawRadius), 2.0f);

		//Engine::GraphicsEngine->GetLineRenderer()->AddPointLocator(VegetationSpots[i], 5.0f);
	}*/

	// Seed randomizer again
	//srand(timeGetTime());

	if(DrawBoundingBox)
		Engine::GraphicsEngine->GetLineRenderer()->AddAABBMinMax(BoxMin, BoxMax);

	Engine::GAPI->GetRendererState()->RasterizerState.CullMode = GothicRasterizerStateInfo::CM_CULL_FRONT;
	Engine::GAPI->GetRendererState()->RasterizerStateDirty = true;

	Engine::GAPI->GetRendererState()->BlendState.AlphaToCoverage = false;
	Engine::GAPI->GetRendererState()->BlendStateDirty = true;
}

/** Sets bounding box rendering */
void GVegetationBox::SetRenderBoundingBox(bool value)
{
	DrawBoundingBox = value;
}