#pragma once
#include "pch.h"
#include "GothicGraphicsState.h"
#include "WorldConverter.h"
#include "zCTree.h"
#include "zTypes.h"

struct ParticleRenderInfo
{
	GothicBlendStateInfo BlendState;
};

struct ParticleInstanceInfo
{
	D3DXMATRIX worldview;
	DWORD color;
	float2 scale;
};

struct BspVobInfo
{
	BspVobInfo()
	{
		NumStaticLights = 0;
	}

	std::list<VobInfo *> Vobs;
	std::list<VobInfo *> IndoorVobs;
	std::list<VobInfo *> SmallVobs;
	std::list<VobLightInfo *> Lights;
	std::list<VobLightInfo *> IndoorLights;
	int NumStaticLights;
};

struct CameraReplacement
{
	D3DXMATRIX ViewReplacement;
	D3DXMATRIX ProjectionReplacement;
	D3DXVECTOR3 PositionReplacement;
	D3DXVECTOR3 LookAtReplacement;
};

/** Version of this struct */
const int MATERIALINFO_VERSION = 1;

struct MaterialInfo
{
	MaterialInfo()
	{
		SpecularIntensity = 0.3f;
		SpecularPower = 60.0f;
		NormalmapStrength = 1.0f;
	}

	

	/** Writes this info to a file */
	void WriteToFile(const std::string& name);

	/** Loads this info from a file */
	void LoadFromFile(const std::string& name);

	float SpecularIntensity;
	float SpecularPower;
	float NormalmapStrength;
	float DisplacementFactor;
};

/** Class used to communicate between Gothic and the Engine */
class zCPolygon;
class zCTexture;
class zCParticleFX;
class zCVisual;
class GSky;
class GMesh;
class zCBspBase;
class GInventory;
class zCVobLight;
class MyDirectDrawSurface7;
class GVegetationBox;
class GothicAPI
{
public:
	GothicAPI(void);
	~GothicAPI(void);

	/** Called when the window got set */
	void OnSetWindow(HWND hWnd);

	/** Message-Callback for the main window */
	LRESULT OnWindowMessage(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

	/** Called when the game loaded a new level */
	void OnGeometryLoaded(zCPolygon** polys, unsigned int numPolygons);

	/** Called when the game is done loading the world */
	void OnWorldLoaded();

	/** Called to update the world, before rendering */
	void OnWorldUpdate();

	/** Called when a VOB got added to the BSP-Tree or the world */
	void OnAddVob(zCVob* vob, zCWorld* world);

	/** Called when a VOB got removed from the world */
	void OnRemovedVob(zCVob* vob, zCWorld* world);

	/** Called when a material got removed */
	void OnMaterialDeleted(zCMaterial* mat);

	/** Called when a particle system got removed */
	void OnParticleFXDeleted(zCParticleFX* pfx);

	/** Called when a visual got removed */
	void OnVisualDeleted(zCVisual* visual);
	

	/** Called when a material got removed */
	void OnMaterialCreated(zCMaterial* mat);

	/** Returns the GraphicsState */
	GothicRendererState* GetRendererState();

	/** Returns in which directory we started in */
	const std::string& GetStartDirectory();

	/** Draws the world-mesh */
	void DrawWorldMeshNaive();

	/** Draws a skeletal mesh-vob */
	void DrawSkeletalMeshVob(SkeletalVobInfo* vi);

	/** Draws the inventory */
	void DrawInventory(zCWorld* world, zCCamera& camera);

	/** Locks the resource CriticalSection */
	void EnterResourceCriticalSection();

	/** Unlocks the resource CriticalSection */
	void LeaveResourceCriticalSection();

	/** Draws a MeshInfo */
	void DrawMeshInfo(zCMaterial* mat, MeshInfo* msh);

	/** Draws a SkeletalMeshInfo */
	void DrawSkeletalMeshInfo(zCMaterial* mat, SkeletalMeshInfo* msh, std::vector<D3DXMATRIX>& transforms);

	/** Draws a zCParticleFX */
	void DrawParticleFX(zCVob* source, zCParticleFX* fx);

	/** Sets the world matrix */
	void SetWorldTransform(const D3DXMATRIX& world);

	/** Sets the world matrix */
	void SetViewTransform(const D3DXMATRIX& view);

	/** Sets the world matrix */
	void SetWorldTransform(const D3DXMATRIX& world, bool transpose);

	/** Sets the world matrix */
	void SetViewTransform(const D3DXMATRIX& view, bool transpose);

	/** Sets the world matrix */
	void SetWorldViewTransform(const D3DXMATRIX& world, const D3DXMATRIX& view);

	/** Sets the world matrix */
	void ResetWorldTransform();

	/** Sets the world matrix */
	void ResetViewTransform();

	/** Debugging */
	void DrawTriangle();

	/** Returns the loaded sections */
	std::map<int, std::map<int, WorldMeshSectionInfo>>& GetWorldSections();

	/** Returns the loaded skeletal mesh vobs */
	std::list<SkeletalVobInfo *>& GetSkeletalMeshVobs();

	/** Returns the current cameraposition */
	D3DXVECTOR3 GetCameraPosition();

	/** Returns the view matrix */
	void GetViewMatrix(D3DXMATRIX* view);

	/** Returns the view matrix */
	void GetInverseViewMatrix(D3DXMATRIX* invView);

	/** Returns the projection-matrix */
	D3DXMATRIX& GetProjectionMatrix();

	/** Unprojects a pixel-position on the screen */
	void Unproject(const D3DXVECTOR3& p, D3DXVECTOR3* worldPos, D3DXVECTOR3* worldDir);

	/** Unprojects the current cursor, returns it's direction in world-space */
	D3DXVECTOR3 UnprojectCursor();

	/** Traces the worldmesh and returns the hit-location */
	bool TraceWorldMesh(const D3DXVECTOR3& origin, const D3DXVECTOR3& dir, D3DXVECTOR3& hit, std::string* hitTextureName = NULL, D3DXVECTOR3* hitTriangle = NULL, MeshInfo** hitMesh = NULL, zCMaterial** hitMaterial = NULL);

	/** Returns the GSky-Object */
	GSky* GetSky();

	/** Returns the inventory */
	GInventory* GetInventory();

	/** Returns if the material is currently active */
	bool IsMaterialActive(zCMaterial* mat);

	/** Sets the current input state */
	void SetEnableGothicInput(bool value);

	/** Returns the midpoint of the current world */
	WorldInfo* GetLoadedWorldInfo(){return LoadedWorldInfo;}

	/** Returns gothics fps-counter */
	int GetFramesPerSecond();

	/** Returns the current frame time */
	float GetFrameTimeSec();

	/** Returns global time */
	float GetTimeSeconds();

	/** Builds the static mesh instancing cache */
	void BuildStaticMeshInstancingCache();

	/** Draws the AABB for the BSP-Tree using the line renderer*/
	void DebugDrawBSPTree();

	/** Recursive helper function to draw the BSP-Tree */
	void DebugDrawTreeNode(zCBspBase* base, zTBBox3D boxCell, int clipFlags = 63);

	/** Collects vobs using gothics BSP-Tree */
	void CollectVisibleVobs(std::list<VobInfo *>& vobs, std::list<VobLightInfo *>& lights);

	/** Collects visible sections from the current camera perspective */
	void CollectVisibleSections(std::list<WorldMeshSectionInfo*>& sections);

	/** Builds our BspTreeVobMap */
	void BuildBspVobMapCache();

	/** Disables a problematic method which causes the game to conflict with other applications on startup */
	static void DisableErrorMessageBroadcast();

	/** Sets/Gets the far-plane */
	void SetFarPlane(float value);
	float GetFarPlane();

	/** Sets/Gets the far-plane */
	void SetNearPlane(float value);
	float GetNearPlane();

	/** Resets the object, like at level load */
	void ResetWorld();

	/** Get material by texture name */
	zCMaterial* GetMaterialByTextureName(const std::string& name);
	void GetMaterialListByTextureName(const std::string& name, std::list<zCMaterial*>& list);

	/** Returns the time since the last frame */
	float GetDeltaTime();

	/** If this returns true, the property holds the name of the currently bound texture. If that is the case, any MyDirectDrawSurfaces should not bind themselfes
		to the pipeline, but rather check if there are additional textures to load */
	bool IsInTextureTestBindMode(std::string& currentBoundTexture);

	/** Sets the current texture test bind mode status */
	void SetTextureTestBindMode(bool enable, const std::string& currentTexture);

	/** Sets the CameraReplacementPtr */
	void SetCameraReplacementPtr(CameraReplacement* ptr){CameraReplacementPtr = ptr;}

	/** Lets Gothic draw its sky */
	void DrawSkyGothicOriginal();

	/** Returns the material info associated with the given material */
	MaterialInfo* GetMaterialInfoFrom(zCTexture* tex);

	/** Adds a surface */
	void AddSurface(const std::string& name, MyDirectDrawSurface7* surface);

	/** Removes a surface */
	void RemoveSurface(MyDirectDrawSurface7* surface);

	/** Resets all vob-stats drawn this frame */
	void ResetVobFrameStats(std::list<VobInfo *>& vobs);

	/** Sets the currently bound texture */
	void SetBoundTexture(int idx, zCTexture* tex);
	zCTexture* GetBoundTexture(int idx);

	/** Returns gothics output window */
	HWND GetOutputWindow(){return OutputWindow;}

	/** Spawns a vegetationbox at the camera */
	GVegetationBox* SpawnVegetationBoxAt(const D3DXVECTOR3& position, 
		const D3DXVECTOR3& min = D3DXVECTOR3(-1000, -500, -1000), 
		const D3DXVECTOR3& max = D3DXVECTOR3(1000, 500, 1000), 
		float density = 1.0f,
		const std::string& restrictByTexture = "");

	/** Adds a vegetationbox to the world */
	void AddVegetationBox(GVegetationBox* box);

private:


	/** Helper function for going through the bsp-tree */
	void BuildBspVobMapCacheHelper(zCBspBase* base);

	/** Recursive helper function to draw collect the vobs */
	void CollectVisibleVobsHelper(zCBspBase* base, zTBBox3D boxCell, int clipFlags, std::list<VobInfo *>& vobs, std::list<VobLightInfo  *>& lights);

	/** Hooked Window-Proc from the game */
	static LRESULT CALLBACK GothicWndProc(
		HWND hWnd,
		UINT msg,
		WPARAM wParam,
		LPARAM lParam
		);


	/** Debugging */
	void DrawTestInstances();

	/** Goes through the given zCTree and registeres all found vobs */
	void TraverseVobTree(zCTree<zCVob>* tree);

	/** Saved Graphics state */
	GothicRendererState RendererState;

	/** Loaded world mitpoint */
	WorldInfo* LoadedWorldInfo;

	/** Currently bound textures from gothic */
	zCTexture* BoundTextures[8];

	std::map<zCTexture*, std::vector<ParticleInstanceInfo>> FrameParticles;
	std::map<zCTexture*, ParticleRenderInfo> FrameParticleInfo;

	/** Loaded game sections */
	std::map<int, std::map<int, WorldMeshSectionInfo>> WorldSections;

	/** List of vobs with skeletal meshes (Having a zCModel-Visual) */
	std::list<SkeletalVobInfo *> SkeletalMeshVobs;

	/** List of Vobs having a zCParticleFX-Visual */
	std::list<zCVob *> ParticleEffectVobs;
	std::hash_map<zCVob *, std::string> tempParticleNames;

	/** Set of Materials */
	std::set<zCMaterial *> LoadedMaterials;

	/** List of meshes rendered for this frame */
	std::set<MeshVisualInfo *> FrameMeshInstances;

	/** Map for static mesh visuals */
	std::map<zCProgMeshProto*, MeshVisualInfo*> StaticMeshVisuals;

	/** Map for skeletal mesh visuals */
	std::map<zCModel*, SkeletalMeshVisualInfo*> SkeletalMeshVisuals;

	/** Set of all vobs we registered by now */
	std::set<zCVob*> RegisteredVobs;

	/** Map of vobs and VobIndfos */
	std::hash_map<zCVob*, VobInfo*> VobMap;
	std::hash_map<zCVobLight*, VobLightInfo*> VobLightMap;
	std::hash_map<zCVob*, SkeletalVobInfo*> SkeletalVobMap;

	/** Map of VobInfo-Lists for zCBspLeafs */
	std::hash_map<zCBspBase *, BspVobInfo> BspLeafVobLists;

	/** Map for the material infos */
	std::hash_map<zCTexture*, MaterialInfo> MaterialInfos;

	/** Map of textures */
	std::hash_map<std::string, MyDirectDrawSurface7*> SurfacesByName;

	/** Directory we started in */
	std::string StartDirectory;

	/** Resource critical section */
	CRITICAL_SECTION ResourceCriticalSection;

	/** Sky renderer */
	GSky* SkyRenderer;

	/** Inventory manager */
	GInventory* Inventory;

	/** Saved Wnd-Proc pointer from the game */
	LONG_PTR OriginalGothicWndProc;

	/** Whether we test texture binds to figure out what surface uses which zCTexture object */
	bool TextureTestBindMode;
	std::string BoundTestTexture;

	/** Replacement values for the camera */
	CameraReplacement* CameraReplacementPtr;

	/** List of available GVegetationBoxes */
	std::list<GVegetationBox*> VegetationBoxes;

	/** Gothics output window */
	HWND OutputWindow;
};

